@@ -1,17 +1,18 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import weakref
 from collections import deque, namedtuple
 from functools import partial
+from contextlib import suppress
 
 from .borders import Borders
 from .child import Child
 from .constants import appname, get_boss, is_macos, is_wayland
 from .fast_data_types import (
-    add_tab, glfw_post_empty_event, mark_tab_bar_dirty, next_window_id,
-    pt_to_px, remove_tab, remove_window, set_active_tab, swap_tabs,
+    add_tab, mark_tab_bar_dirty, next_window_id,
+    pt_to_px, remove_tab, remove_window, ring_bell, set_active_tab, swap_tabs,
     x11_window_id
 )
 from .layout import create_layout_object_for, evict_cached_layouts
@@ -27,6 +28,14 @@ def SpecialWindow(cmd, stdin=None, override_title=None, cwd_from=None, cwd=None,
     return SpecialWindowInstance(cmd, stdin, override_title, cwd_from, cwd, overlay_for, env)
 
 
+def add_active_id_to_history(items, item_id, maxlen=64):
+    with suppress(ValueError):
+        items.remove(item_id)
+    items.append(item_id)
+    if len(items) > maxlen:
+        items.popleft()
+
+
 class Tab:  # {{{
 
     def __init__(self, tab_manager, session_tab=None, special_window=None, cwd_from=None):
@@ -34,6 +43,7 @@ class Tab:  # {{{
         self.tab_manager_ref = weakref.ref(tab_manager)
         self.os_window_id = tab_manager.os_window_id
         self.id = add_tab(self.os_window_id)
+        self.active_window_history = deque()
         if not self.id:
             raise Exception('No OS window with id {} found, or tab counter has wrapped'.format(self.os_window_id))
         self.opts, self.args = tab_manager.opts, tab_manager.args
@@ -84,6 +94,9 @@ class Tab:  # {{{
             old_active_window = self.windows[self._active_window_idx]
         except Exception:
             old_active_window = None
+        else:
+            wid = old_active_window.id if old_active_window.overlay_for is None else old_active_window.overlay_for
+            add_active_id_to_history(self.active_window_history, wid)
         self._active_window_idx = max(0, min(val, len(self.windows) - 1))
         try:
             new_active_window = self.windows[self._active_window_idx]
@@ -185,6 +198,16 @@ class Tab:  # {{{
             return
         return 'Could not resize'
 
+    def resize_window(self, quality, increment):
+        if increment < 1:
+            raise ValueError(increment)
+        is_horizontal = quality in ('wider', 'narrower')
+        increment *= 1 if quality in ('wider', 'taller') else -1
+        if self.resize_window_by(
+                self.windows[self.active_window_idx].id,
+                increment, is_horizontal) is not None:
+            ring_bell()
+
     def reset_window_sizes(self):
         if self.current_layout.remove_all_biases():
             self.relayout()
@@ -199,7 +222,7 @@ class Tab:  # {{{
         if env:
             fenv.update(env)
         fenv['KITTY_WINDOW_ID'] = str(next_window_id())
-        if not is_macos and not is_wayland:
+        if not is_macos and not is_wayland():
             try:
                 fenv['WINDOWID'] = str(x11_window_id(self.os_window_id))
             except Exception:
@@ -220,7 +243,6 @@ class Tab:  # {{{
         get_boss().add_child(window)
         self.active_window_idx = self.current_layout.add_window(self.windows, window, self.active_window_idx)
         self.relayout_borders()
-        glfw_post_empty_event()
         return window
 
     def new_special_window(self, special_window):
@@ -230,17 +252,42 @@ class Tab:  # {{{
         if self.windows:
             self.remove_window(self.windows[self.active_window_idx])
 
+    def previous_active_window_idx(self, num):
+        try:
+            old_window_id = self.active_window_history[-num]
+        except IndexError:
+            return
+        for idx, w in enumerate(self.windows):
+            if w.id == old_window_id:
+                return idx
+
     def remove_window(self, window):
-        self.active_window_idx = self.current_layout.remove_window(self.windows, window, self.active_window_idx)
+        idx = self.previous_active_window_idx(1)
+        next_window_id = None
+        if idx is not None:
+            next_window_id = self.windows[idx].id
+        active_window_idx = self.current_layout.remove_window(self.windows, window, self.active_window_idx)
         remove_window(self.os_window_id, self.id, window.id)
+        if next_window_id is None and active_window_idx is not None and len(self.windows) > active_window_idx:
+            next_window_id = self.windows[active_window_idx].id
+        if next_window_id is not None:
+            for idx, window in enumerate(self.windows):
+                if window.id == next_window_id:
+                    self.active_window_idx = self.current_layout.set_active_window(self.windows, idx)
+                    break
+            else:
+                self.active_window_idx = active_window_idx
+        else:
+            self.active_window_idx = active_window_idx
         self.relayout_borders()
-        glfw_post_empty_event()
+        active_window = self.active_window
+        if active_window:
+            self.title_changed(active_window)
 
     def set_active_window_idx(self, idx):
         if idx != self.active_window_idx:
             self.active_window_idx = self.current_layout.set_active_window(self.windows, idx)
             self.relayout_borders()
-            glfw_post_empty_event()
 
     def set_active_window(self, window):
         try:
@@ -255,15 +302,19 @@ class Tab:  # {{{
 
     def nth_window(self, num=0):
         if self.windows:
-            self.active_window_idx = self.current_layout.nth_window(self.windows, num)
+            if num < 0:
+                idx = self.previous_active_window_idx(-num)
+                if idx is None:
+                    return
+                self.active_window_idx = self.current_layout.set_active_window(self.windows, idx)
+            else:
+                self.active_window_idx = self.current_layout.nth_window(self.windows, num)
             self.relayout_borders()
-            glfw_post_empty_event()
 
     def _next_window(self, delta=1):
         if len(self.windows) > 1:
             self.active_window_idx = self.current_layout.next_window(self.windows, self.active_window_idx, delta)
             self.relayout_borders()
-            glfw_post_empty_event()
 
     def next_window(self):
         self._next_window()
@@ -279,12 +330,10 @@ class Tab:  # {{{
         if candidates:
             self.active_window_idx = self.current_layout.set_active_window(self.windows, candidates[0])
             self.relayout_borders()
-            glfw_post_empty_event()
 
     def move_window(self, delta=1):
         self.active_window_idx = self.current_layout.move_window(self.windows, self.active_window_idx, delta)
         self.relayout()
-        glfw_post_empty_event()
 
     def move_window_to_top(self):
         self.move_window(-self.active_window_idx)
@@ -330,7 +379,9 @@ class TabManager:  # {{{
 
     def __init__(self, os_window_id, opts, args, startup_session):
         self.os_window_id = os_window_id
+        self.last_active_tab_id = None
         self.opts, self.args = opts, args
+        self.tab_bar_hidden = self.opts.tab_bar_style == 'hidden'
         self.tabs = []
         self.active_tab_history = deque()
         self.tab_bar = TabBar(self.os_window_id, opts)
@@ -348,11 +399,10 @@ class TabManager:  # {{{
     def active_tab_idx(self, val):
         try:
             old_active_tab = self.tabs[self._active_tab_idx]
-            self.active_tab_history.append(old_active_tab.id)
-            if len(self.active_tab_history) > 64:
-                self.active_tab_history.popleft()
         except Exception:
             old_active_tab = None
+        else:
+            add_active_id_to_history(self.active_tab_history, old_active_tab.id)
         self._active_tab_idx = max(0, min(val, len(self.tabs) - 1))
         try:
             new_active_tab = self.tabs[self._active_tab_idx]
@@ -369,19 +419,24 @@ class TabManager:  # {{{
                     w.focus_changed(True)
 
     def refresh_sprite_positions(self):
-        self.tab_bar.screen.refresh_sprite_positions()
+        if not self.tab_bar_hidden:
+            self.tab_bar.screen.refresh_sprite_positions()
+
+    @property
+    def tab_bar_should_be_visible(self):
+        return len(self.tabs) >= self.opts.tab_bar_min_tabs
 
     def _add_tab(self, tab):
-        before = len(self.tabs)
+        visible_before = self.tab_bar_should_be_visible
         self.tabs.append(tab)
-        if len(self.tabs) > 1 and before < 2:
+        if not visible_before and self.tab_bar_should_be_visible:
             self.tabbar_visibility_changed()
 
     def _remove_tab(self, tab):
-        before = len(self.tabs)
+        visible_before = self.tab_bar_should_be_visible
         remove_tab(self.os_window_id, tab.id)
         self.tabs.remove(tab)
-        if len(self.tabs) < 2 and before > 1:
+        if visible_before and not self.tab_bar_should_be_visible:
             self.tabbar_visibility_changed()
 
     def _set_active_tab(self, idx):
@@ -389,12 +444,12 @@ class TabManager:  # {{{
         set_active_tab(self.os_window_id, idx)
 
     def tabbar_visibility_changed(self):
-        self.tab_bar.layout()
-        self.resize(only_tabs=True)
-        glfw_post_empty_event()
+        if not self.tab_bar_hidden:
+            self.tab_bar.layout()
+            self.resize(only_tabs=True)
 
     def mark_tab_bar_dirty(self):
-        if len(self.tabs) > 1:
+        if self.tab_bar_should_be_visible and not self.tab_bar_hidden:
             mark_tab_bar_dirty(self.os_window_id)
 
     def update_tab_bar_data(self):
@@ -402,8 +457,9 @@ class TabManager:  # {{{
 
     def resize(self, only_tabs=False):
         if not only_tabs:
-            self.tab_bar.layout()
-            self.mark_tab_bar_dirty()
+            if not self.tab_bar_hidden:
+                self.tab_bar.layout()
+                self.mark_tab_bar_dirty()
         for tab in self.tabs:
             tab.relayout()
 
@@ -424,8 +480,17 @@ class TabManager:  # {{{
             self.set_active_tab_idx((self.active_tab_idx + len(self.tabs) + delta) % len(self.tabs))
 
     def goto_tab(self, tab_num):
-        if tab_num < len(self.tabs) and 0 <= tab_num:
+        if 0 <= tab_num < len(self.tabs):
             self.set_active_tab_idx(tab_num)
+        elif tab_num < 0:
+            try:
+                old_active_tab_id = self.active_tab_history[tab_num]
+            except IndexError:
+                return
+            for idx, tab in enumerate(self.tabs):
+                if tab.id == old_active_tab_id:
+                    self.set_active_tab_idx(idx)
+                    break
 
     def __iter__(self):
         return iter(self.tabs)
@@ -439,6 +504,7 @@ class TabManager:  # {{{
                 'id': tab.id,
                 'is_focused': tab is active_tab,
                 'title': tab.name or tab.title,
+                'layout': tab.current_layout.name,
                 'windows': list(tab.list_windows(active_window)),
             }
 
@@ -461,8 +527,10 @@ class TabManager:  # {{{
         if len(self.tabs) > 1:
             idx = self.active_tab_idx
             nidx = (idx + len(self.tabs) + delta) % len(self.tabs)
-            self.tabs[idx], self.tabs[nidx] = self.tabs[nidx], self.tabs[idx]
-            swap_tabs(self.os_window_id, idx, nidx)
+            step = 1 if idx < nidx else -1
+            for i in range(idx, nidx, step):
+                self.tabs[i], self.tabs[i + step] = self.tabs[i + step], self.tabs[i]
+                swap_tabs(self.os_window_id, i, i + step)
             self._set_active_tab(nidx)
             self.mark_tab_bar_dirty()
 
@@ -472,8 +540,9 @@ class TabManager:  # {{{
         self._add_tab(Tab(self, special_window=special_window, cwd_from=cwd_from))
         self._set_active_tab(idx)
         if len(self.tabs) > 2 and as_neighbor and idx != nidx:
-            self.tabs[idx], self.tabs[nidx] = self.tabs[nidx], self.tabs[idx]
-            swap_tabs(self.os_window_id, idx, nidx)
+            for i in range(idx, nidx, -1):
+                self.tabs[i], self.tabs[i-1] = self.tabs[i-1], self.tabs[i]
+                swap_tabs(self.os_window_id, i, i-1)
             self._set_active_tab(nidx)
             idx = nidx
         self.mark_tab_bar_dirty()
@@ -482,16 +551,25 @@ class TabManager:  # {{{
     def remove(self, tab):
         self._remove_tab(tab)
         next_active_tab = -1
-        while self.active_tab_history and next_active_tab < 0:
-            tab_id = self.active_tab_history.pop()
-            if tab_id == tab.id:
-                continue
-            for idx, qtab in enumerate(self.tabs):
-                if qtab.id == tab_id:
-                    next_active_tab = idx
-                    break
+        while True:
+            try:
+                self.active_tab_history.remove(tab.id)
+            except ValueError:
+                break
+
+        if self.opts.tab_switch_strategy == 'previous':
+            while self.active_tab_history and next_active_tab < 0:
+                tab_id = self.active_tab_history.pop()
+                for idx, qtab in enumerate(self.tabs):
+                    if qtab.id == tab_id:
+                        next_active_tab = idx
+                        break
+        elif self.opts.tab_switch_strategy == 'left':
+            next_active_tab = max(0, self.active_tab_idx - 1)
+
         if next_active_tab < 0:
             next_active_tab = max(0, min(self.active_tab_idx, len(self.tabs) - 1))
+
         self._set_active_tab(next_active_tab)
         self.mark_tab_bar_dirty()
         tab.destroy()
@@ -517,7 +595,7 @@ class TabManager:  # {{{
 
     @property
     def blank_rects(self):
-        return self.tab_bar.blank_rects if len(self.tabs) > 1 else ()
+        return self.tab_bar.blank_rects if self.tab_bar_should_be_visible else ()
 
     def destroy(self):
         for t in self:
